/***************************************************************************
 *	Copyright (C) 2005 by Karye												*
 *	karye@users.sourceforge.net												*
 *																			*
 *	This program is free software; you can redistribute it and/or modify	*
 *	it under the terms of the GNU General Public License as published by	*
 *	the Free Software Foundation; either version 2 of the License, or		*
 *	(at your option) any later version.										*
 *																			*
 *	This program is distributed in the hope that it will be useful,			*
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of			*
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the			*
 *	GNU General Public License for more details.							*
 *																			*
 *	You should have received a copy of the GNU General Public License		*
 *	along with this program; if not, write to the							*
 *	Free Software Foundation, Inc.,											*
 *	59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.				*
 ***************************************************************************/

#include <qdebug.h>       // for QDebug
#include <qdir.h>         // for QDir, operator|, QDir::Name, QDir::NoSymLinks
#include <qglobal.h>      // for qWarning, foreach, qCritical
#include <qlist.h>        // for QList<>::iterator, QList<>::Iterator, QList
#include <qmap.h>         // for QMap, QMap<>::iterator
#include <qobjectdefs.h>  // for Q_EMIT
#include <qstring.h>      // for QString, operator+, operator<
#include <qstringlist.h>  // for QStringList

#include "cacheportagejob.h"
#include "common.h"       // for KurooDBSingleton, SignalistSingleton, DEBUG...
#include "global.h"       // for parsePackage
#include "portagedb.h"    // for KurooDB, DbConnection
#include "progressjob.h"  // for ProgressJob
#include "settings.h"     // for KurooConfig
#include "signalist.h"    // for Signalist

namespace ThreadWeaver {
class Thread;
}  // namespace ThreadWeaver

/**
 * @class CachePortageJob
 * @short Thread to cache package information from the Portage directory to speed up portage refreshing.
 *
 * Portage cache is scanned for package sizes, and stored in portage cache map and in the database.
 */
CachePortageJob::CachePortageJob()
    : m_db( KurooDBSingleton::Instance()->getStaticDbConnection() )
{
}

CachePortageJob::~CachePortageJob()
{
	KurooDBSingleton::Instance()->returnStaticDbConnection( m_db );
}

/**
 * Scan for package size found in digest files.
 */
void CachePortageJob::run( ThreadWeaver::JobPointer  /*self*/, ThreadWeaver::Thread*  /*thread*/)
{
	DEBUG_LINE_INFO;
	if ( !m_db->isConnected() ) {
		qCritical() << "Creating cache. Can not connect to database";
		return;// false;
	}

	int count( 0 );
	QMap <QString, uint> packageNameToSizeInBytes;
	QDir dCategory, dPackage;
	dCategory.setFilter( QDir::Dirs | QDir::NoSymLinks | QDir::NoDotAndDotDot );
	QStringList nonCategories = { QStringLiteral("distfiles"), QStringLiteral("eclass"), QStringLiteral("licenses"), QStringLiteral("metadata"), QStringLiteral("packages"), QStringLiteral("profiles"), QStringLiteral("scripts") };	// for filtering out false errors
	dPackage.setFilter( QDir::NoDotAndDotDot );
	//dCategory.setSorting( QDir::Name );

	// Get a count of total packages for proper progress
	int packageCount = KurooDBSingleton::Instance()->singleQuery( QStringLiteral("SELECT data FROM dbInfo WHERE meta = 'packageCount' LIMIT 1;"), m_db ).toInt();
	setTotalSteps( packageCount == 0 ? 35000 : packageCount );
	//TODO: setStatus( "CachePortage", i18n("Collecting package information...") );

	//TODO: kuroo was already built to handle overlays?
	// Get list of categories in Portage and Overlays
	QStringList pathList = QString(KurooConfig::dirPortage() + u' ' + KurooConfig::dirPortageOverlay()).split(u' ', Qt::SkipEmptyParts);

	// Scan Portage cache
	for (auto & itPath : pathList) {
		if ( !dCategory.cd( itPath ) ) {
			qWarning() << "Creating cache. Can not access " << itPath;
			continue;
		}

		for ( auto & itCategory : dCategory.entryList() ) {
			//It's probably fewer cycles to just allow special non-categories error and filter out false-negative errors later

            // Abort the scan
			if ( isAborted() ) {
				qWarning() << "Creating cache. Aborted!";
				//TODO: setStatus( "CachePortage", i18n("Caching aborted.") );
				//Q_EMIT failed( self );
				setStatus( ThreadWeaver::JobInterface::Status::Status_Failed );
				SignalistSingleton::Instance()->cachePortageComplete();
				return;
            }

			// Get list of package-versions in this category
			QString currentManifest;
			QHash<QString, uint> manifestFilesToSize;
			dPackage.setFilter( QDir::Files | QDir::NoSymLinks | QDir::NoDotAndDotDot );
			dPackage.setSorting( QDir::Name );
			if ( dPackage.cd( itPath + QStringLiteral("/metadata/md5-cache/") + itCategory ) ) {
				for ( auto & itPackage : dPackage.entryList() ) {
					if ( itPackage == QStringLiteral("Manifest.gz") )
						continue;

                    // Abort the scan
					if ( isAborted() ) {
						qWarning() << "Creating cache. Aborted!";
						//TODO: setStatus( "CachePortage", i18n("Caching aborted.") );
						setStatus( ThreadWeaver::JobInterface::Status::Status_Failed );
						//Q_EMIT failed( self );
						SignalistSingleton::Instance()->cachePortageComplete();
						return;
                    }
                    //Includes version
					QString package = itCategory + u'/' + itPackage;

                    QStringList parts = parsePackage( itPackage );
					if ( parts.isEmpty() ) {
						qWarning() << __PRETTY_FUNCTION__ << "Can't parse" << itPackage;
						continue;
					}

					QString packageName = parts[1];
					QString newManifest( itPath + u'/' + itCategory + u'/' + packageName + QStringLiteral("/Manifest") ); 
					if ( currentManifest != newManifest ) {
						manifestFilesToSize.clear();
						currentManifest = newManifest;
						QFile manifestFile( currentManifest );
						if ( manifestFile.open( QIODevice::ReadOnly ) ) {
							QTextStream lines( &manifestFile );
							while ( !lines.atEnd() ) {
								auto tokens = lines.readLine().split(u' ');	//No way to limit the number of tokens?
								//each line is <type> <name> <size> (<checksum> <checksumtype>)+ per https://www.gentoo.org/glep/glep-0044.html
								manifestFilesToSize[tokens[1]] = tokens[2].toUInt();	//Returns 0 if conversion failed, would be nice to qWarn
							}
						}
						else {
							qWarning() << __PRETTY_FUNCTION__ << "Can't open" << currentManifest;
						}
					}

					//TODO: /files/digest-* doesn't exist anymore, this will have to combine /metadata/md5-cache/<cat>/<pkg>-<version> with 'Manifest'
					// Get package size
					//HACK: packageNameToSizeInBytes.insert( package, QStringLiteral("12345") );
					/*QString path = itPath + "/" + itCategory + "/" + packageName + "/files/digest-" + itPackage;
					QFile file( path );
					if ( file.open( QIODevice::ReadOnly ) ) {
						//Digests were named /files/<package>-<version> and used to look like:
// 							MD5 3c178889822836378fcf526195b86814 /files/kdelibs-3.1.1-kio_http.diff 5257
// 							MD5 24190dc0508cd29ea5adecc4cefc0db2 /files/kdelibs-2.2.2a-crosside.diff 1267
// 							MD5 b9e995da45022f78aed48a79f02201e7 /files/kdelibs-3.1-bookmark.diff 1320
// 							MD5 acd90863c0847e230aa37f3943d58176 /files/kdelibs-3.0.5b-libxml2-2.5.2.diff 703
// 							MD5 41d30ba8cf8d36ad3caebeedb1ddfd0e /files/kdelibs-3.1.1-kjs-alphaev6-gcc3-workaround.patch 477
// 							MD5 7353213c0eaadb42e684fb98fe8f4ff0 /files/kdelibs-3.1.1-zip.diff 2906
// 							MD5 7353213c0eaadb42e684fb98fe8f4ff0 /files/kdelibs-3.1-zip.diff 2906
// 							MD5 41d30ba8cf8d36ad3caebeedb1ddfd0e /files/kdelibs-3.1.1a-kjs-alphaev6-gcc3-workaround.patch 477
// 							MD5 41d30ba8cf8d36ad3caebeedb1ddfd0e /files/kdelibs-3.1-kjs-alphaev6-gcc3-workaround.patch 477
// 							MD5 bcd0e223c0be9cc4d1af5f2b5578a096 /files/kdelibs-3.0.5a-dontcrash.diff 500
// 							MD5 ed6c2a4ca2dca736f3cbfe717b762fe6 /files/kdelibs-2.2.2a-gentoo.diff 75788
// 							MD5 6e7c2097ad93ecd9a011d1685cb8971f /files/kdelibs-3.1-libxml2-2.5.2.diff 1404
// 							MD5 acd90863c0847e230aa37f3943d58176 /files/kdelibs-3.0.5a-libxml2-2.5.2.diff 703
// 							MD5 6a361c265a042beaa15c060bf8ce4057 /kdelibs-3.0.5b.ebuild 1893
// 							MD5 50b483665bc868f2dbc53aaaa3c2f302 kdelibs-3.0.5b.tar.bz2 7610088
						// based on NeddySeagon's portage snapshot at https://wiki.gentoo.org/wiki/User:NeddySeagoon/Historical_Gentoo
						std::ifstream in( path );
						std::string word;
						while ( in >> word );
						packageNameToSizeInBytes.insert( package, word );
						file.close();
					}
					else
						qWarning() << "Creating cache. Reading: " << path;*/
					//metadata/md5-cache/<category>/<package>-<version>
					auto path = dPackage.absolutePath() + u'/' + itPackage;
					QFile file( path );
					if ( file.open( QIODevice::ReadOnly ) ) {
						//This is the md5-cache file with all vars listed in dependency specification syntax per https://projects.gentoo.org/pms/7/pms.html#x1-68002r2
						uint runningSize = 0;
						QTextStream lines( &file );
						while ( !lines.atEnd() ) {
							auto line = lines.readLine();
							if ( line.startsWith( QStringLiteral("SRC_URI=") ) ) {
								//Chomp off 'SRC_URI=' and tokenize by spaces
								auto sources = line.mid( 8 ).split(u' ');
								// For any construct like "<URL> -> <filename>" the URL is just renamed to filename and manifested as filename, so ignore '->' and 'URL'
								while ( int i = sources.lastIndexOf(QStringLiteral("->"), i) ) {
									if ( i == -1 )
										break;
									sources.removeAt(i);
									sources.removeAt(--i);
									i = sources.lastIndexOf(QStringLiteral("->"), i);
								}
								//Gather all required distfiles, ignoring anything inside parens
								int conditionalStack = 0;
								for ( auto token : sources ) {
									if ( token == u'(' )
										conditionalStack++;
									else if ( token == u')' )
										conditionalStack--;
									else if ( 0 == conditionalStack) {
										if ( token.contains( u'/' ) )
											token = token.mid( token.lastIndexOf( u'/' ) + 1 );

										if ( manifestFilesToSize.contains( token ) )
											runningSize += manifestFilesToSize.value( token );
// 										else if ( token == QStringLiteral("download") )
// 											qWarning() << __PRETTY_FUNCTION__ << "download came out of" << line;
										else if ( !token.endsWith( u'?' ) )
											qWarning() << __PRETTY_FUNCTION__ << "Couldn't find" << token;
									}
								}
								break;	//no other lines can be SRC_URI= so we can stop now
							}
						}
						packageNameToSizeInBytes.insert( package, runningSize );
					}
					else {
						qDebug() << __PRETTY_FUNCTION__ << __LINE__ << "couldn't read" << path;
					}

					// Post scan count progress
					if ( (++count % 100) == 0 )
						setProgress( count );
				}
			}
			else if ( !nonCategories.contains( itCategory ) )
				qWarning() << __PRETTY_FUNCTION__ << "Can't access " << itPath << "/metadata/cache/" << itCategory;

		}
	}
	KurooDBSingleton::Instance()->query( QStringLiteral("UPDATE dbInfo SET data = '%1' WHERE meta = 'packageCount';")
	                                     .arg( count ), m_db );

	// Store cache in DB
	KurooDBSingleton::Instance()->query( QStringLiteral("DELETE FROM cache;"), m_db );
	KurooDBSingleton::Instance()->query( QStringLiteral("BEGIN TRANSACTION;"), m_db );
	qDebug() << __PRETTY_FUNCTION__ << "inserting" << packageNameToSizeInBytes.size() << "packages into cache";
	for ( const auto& package : packageNameToSizeInBytes.keys() )
		KurooDBSingleton::Instance()->insert( QStringLiteral("INSERT INTO cache (package, size) VALUES ('%1', '%2');").
		                                      arg( package, QString::number( packageNameToSizeInBytes.value(package) ) ), m_db );

	KurooDBSingleton::Instance()->query(QStringLiteral("COMMIT TRANSACTION;"), m_db );

	//TODO: setStatus( "CachePortage", i18n("Done.") );
	setProgress( 0 );
	DEBUG_LINE_INFO;
// 	return true;
// }
//
// void CachePortageJob::completeJob()
// {
	SignalistSingleton::Instance()->cachePortageComplete();
}

